﻿using System;
using System.Drawing;
using System.Windows.Forms;
using Cybele.Thinfinity;

namespace matryoshka
{
    public partial class Form1 : Form
    {

        private VirtualUI vui = new VirtualUI();
        private IJSObject mRemoteLayout = null;

        public Form1()
        {
            InitializeComponent();
            if (vui.Active)
            {
                // Code executed only when the application runs with Thinfinity VirtualUI
                // Sets Google Maps URL to run embedded in an iframe
                txtURL.Text = "https://www.google.com/maps/embed";
            }
            else
            {
                //* Code executed only when the application runs from desktop 
                // Sets Google Maps URL to run in the embedded browser
                txtURL.Text = "https://maps.google.com";
            }
        }

        private void btnGo_Click(object sender, EventArgs e)
        {
            Navigate();
        }

        private void txtURL_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (Convert.ToInt32(e.KeyChar) == 13)
            {
                Navigate();
            }
        }

        private void Navigate()
        {
            string sURL = "";
            if (txtURL.Text.Equals(""))
            {
                sURL = "about:_blank";
            }
            else
            {
                sURL = txtURL.Text;
                if (!sURL.StartsWith("http"))
                {
                    sURL = "http://" + sURL;
                }
            }
            if (vui.Active)
            {
                mRemoteLayout.Properties["url"].AsString = sURL;
            }
            else
            {
                webBrowser1.Navigate(new Uri(sURL));
            }
        }
        private void Form1_Shown(object sender, EventArgs e)
        {
            // Creates jsRO object for iframe control
            mRemoteLayout = new JSObject("layout");
            mRemoteLayout.Properties.Add("windowId").AsString = String.Format("virtualui_canvas_{0}", Handle);
            mRemoteLayout.Properties.Add("url").AsString = txtURL.Text;
            mRemoteLayout.Properties.Add("bounds")
                .OnGet(new JSBinding(        // Adds a "getter" to bounds
                    delegate(IJSObject AObj, IJSProperty AProp)
                    {
                        // Returns a JSON object
                        Point p = pnlNav.PointToScreen(new Point(0, 0));
                        p.Offset(-Left, -Top);
                        AProp.AsJSON = "{" + String.Format("\"left\":{0},\"top\":{1},\"width\":{2},\"height\":{3}",
                                                p.X, p.Y, pnlNav.Width, pnlNav.Height) + "}";
                    }));
            mRemoteLayout.ApplyModel();
        }

        private void pnlNav_Resize(object sender, EventArgs e)
        {
            // Updates all mRemoteLayout's properties "getters"
            if (mRemoteLayout != null)
            {
                mRemoteLayout.ApplyChanges();
            }
        }
    }
}