﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;
using System.Text;
using System.Threading;

namespace Cybele.Thinfinity
{
    /*
     * *********************************************************************************************
     *  Thinfinity_TLB translation
     * *********************************************************************************************
     */

    [Guid("4B85F81B-72A2-4FCD-9A6B-9CAC24B7A511"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch), ComImport]
    public interface IVirtualUI
    {
        bool Start(int timeout);
        void Stop();
        void DownloadFile(string LocalFilename);
        void DownloadFile(string LocalFilename, string RemoteFilename);
        void DownloadFile(string LocalFilename, string RemoteFilename, string MimeType);
        void PrintPdf(string FileName);
        void OpenLinkDlg(string Url, string Caption);
        void SendMessage(string Data);
        void AllowExecute(string Filename);
        void SetImageQualityByWnd(long Wnd, string Classname, int Quality);
        bool Active { get; }
        bool Enabled { get; set; }
        bool DevMode { get; set; }
        bool StdDialogs { get; set; }
        IBrowserInfo BrowserInfo { get; }
        IDevServer DevServer { get; }
        IClientSettings ClientSettings { get; }
    }

    [ComVisible(true), Guid("1C5700BC-2317-4062-B614-0A4E286CFE68"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IEvents
    {
        [DispId(101)]
        void OnGetUploadDir(ref string Directory, ref bool Handled);
        [DispId(102)]
        void OnBrowserResize(ref int Width, ref int Height, ref bool ResizeMaximized);
        [DispId(103)]
        void OnClose();
        [DispId(104)]
        void OnReceiveMessage(string Data);
        [DispId(105)]
        void OnDownloadEnd(string Filename);
    }

    [Guid("4D9F5347-460B-4275-BDF2-F2738E7F6757"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IBrowserInfo
    {
        int ViewWidth { get; set; }
        int ViewHeight { get; set; }
        int BrowserWidth { get; }
        int BrowserHeight { get; }
        int ScreenWidth { get; }
        int ScreenHeight { get; }
        string Username { get; }
        string IPAddress { get; }
        string UserAgent { get; }
        int ScreenResolution { get; }
        BrowserOrientation Orientation { get; }
        string UniqueBrowserId { get; }
        string Location { get; }
    }

    [Guid("B3EAC0CA-D7AB-4AB1-9E24-84A63C8C3F80"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IDevServer
    {
        bool Enabled { get; set; }
        int Port { get; set; }
        bool StartBrowser { get; set; }
    }

    [Guid("439624CA-ED33-47BE-9211-91290F29584A"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IClientSettings
    {
        MouseMoveGestureStyle MouseMoveGestureStyle { get; set; }
        MouseMoveGestureAction MouseMoveGestureAction { get; set; }
        bool CursorVisible { get; set; }
    }

    public enum IJSDataType
    {
        JSDT_NULL = 0,
        JSDT_STRING = 1,
        JSDT_INT = 2,
        JSDT_BOOL = 3,
        JSDT_FLOAT = 4,
        JSDT_JSON = 5
    }

    public enum BrowserOrientation
    {
        PORTRAIT = 0,
        LANDSCAPE = 1
    }

    public enum MouseMoveGestureStyle
    {
        MM_STYLE_RELATIVE = 0,
        MM_STYLE_ABSOLUTE = 1
    }

    public enum MouseMoveGestureAction
    {
        MM_ACTION_MOVE = 0,
        MM_ACTION_WHEEL = 1
    }

    [Guid("6DE2E6A0-3C3A-47DC-9A93-928135EDAC90"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSValue
    {
        IJSDataType DataType { set; get; }
        object RawValue { set; get; }
        string AsString { set; get; }
        int AsInt { set; get; }
        bool AsBool { set; get; }
        float AsFloat { set; get; }
        string AsJSON { set; get; }
    }

    [Guid("E492419B-00AC-4A91-9AE9-9A82B07E89AE"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSNamedValue : IJSValue
    {
        string Name { set; get; }
    }

    [Guid("59342310-79A7-4B14-8B63-6DF05609AE30"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSObject
    {
        void FireEvent(string Name, IJSArguments Arguments);
        void ApplyChanges();
        void ApplyModel();
        string Id { get; set; }
        IJSProperties Properties { get; }
        IJSMethods Methods { get; }
        IJSEvents Events { get; }
        IJSObjects Objects { get; }

        event EventHandler<JSExecuteMethodEventArgs> OnExecuteMethod;
        event EventHandler<JSPropertyChangeEventArgs> OnPropertyChange;
    }

    [ComVisible(true), Guid("A3D640E8-CD18-4196-A1A2-C87C82B0F88B"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSObjectEvents
    {
        [DispId(101)]
        void OnExecuteMethod(IJSObject Sender, IJSMethod Method);
        [DispId(102)]
        void OnPropertyChange(IJSObject Sender, IJSProperty Prop);
    }

    [Guid("C2406011-568E-4EAC-B95C-EF29E4806B86"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSObjects
    {
        int Count { get; }
        IJSObject this[object Index] { get; }
        //IJSObject Find(string Id);
        void Clear();
        IJSObject Add(string Id);
        void Remove(IJSObject Item);
    }

    [Guid("1F95C0E9-E7BF-48C9-AA35-88AD0149109B"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSProperty : IJSNamedValue
    {
        IJSProperty OnGet(IJSBinding Binding);
        IJSProperty OnSet(IJSBinding Binding);
    }

    [Guid("FCBB688F-8FB2-42C1-86FC-0AAF3B2A500C"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSProperties
    {
        int Count { get; }
        IJSProperty this[object Index] { get; }
        //IJSProperty Find(string Name);
        void Clear();
        IJSProperty Add(string Name);
        void Remove(IJSProperty Item);
    }

    [Guid("8F8C4462-D7B5-4696-BAD5-16DFAA6E2601"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSArgument : IJSNamedValue
    {
    }

    [Guid("FC097EF5-6D8A-4C80-A2AD-382FDC75E901"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSArguments
    {
        int Count { get; }
        IJSArgument this[object Index] { get; }
        //IJSArgument Find(string Name);
        void Clear();
        IJSArgument Add(string Name);
        void Remove(IJSArgument Item);
    }

    [Guid("C45D6A8F-AD4A-47BB-AC3A-C125D6D5D27E"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSMethod
    {
        IJSMethod AddArgument(string Name, IJSDataType DataType);
        IJSMethod OnCall(IJSCallback Callback);
        string Name { set; get; }
        IJSArguments Arguments { get; }
        IJSValue ReturnValue { get; }
    }

    [Guid("E4CB461F-586E-4121-ABD7-345B87BC423A"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSMethods
    {
        int Count { get; }
        IJSMethod this[object Index] { get; }
        //IJSMethod Find(string Name);
        void Clear();
        IJSMethod Add(string Name);
        void Remove(IJSMethod Item);
    }

    [Guid("8B66EACD-9619-43CF-9196-DCDA17F5500E"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSEvent
    {
        IJSEvent AddArgument(string Name, IJSDataType DataType);
        IJSEvent ArgumentAsNull(object Index);
        IJSEvent ArgumentAsString(object Index, string Value);
        IJSEvent ArgumentAsInt(object Index, int Value);
        IJSEvent ArgumentAsBool(object Index, bool Value);
        IJSEvent ArgumentAsFloat(object Index, float Value);
        IJSEvent ArgumentAsJSON(object Index, string Value);
        void Fire();
        string Name { set; get; }
        IJSArguments Arguments { get; }

    }

    [Guid("6AE952B3-B6DA-4C81-80FF-D0A162E11D02"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSEvents
    {
        int Count { get; }
        IJSEvent this[object Index] { get; }
        //IJSEvent Find(string Name);
        void Clear();
        IJSEvent Add(string Name);
        void Remove(IJSEvent Item);
    }

    [ComVisible(true), Guid("ACFC2953-37F1-479E-B405-D0BB75E156E6"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSBinding
    {
        void Set(IJSObject Parent, IJSProperty Prop);
    }

    [ComVisible(true), Guid("ADD570A0-491A-4E40-8120-57B4D1245FD3"), InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IJSCallback
    {
        void Callback(IJSObject Parent, IJSMethod Method);
    }


    /*
     * *********************************************************************************************
     *  VirtualUILibrary. Base class which loads the Thinfinity.VirtualUI.dll
     * *********************************************************************************************
     */

    public class VirtualUILibrary
    {
        [DllImport("kernel32.dll")]
        private static extern IntPtr LoadLibrary(string dllToLoad);
        [DllImport("kernel32.dll")]
        protected static extern IntPtr GetProcAddress(IntPtr hModule, string procedureName);
        [DllImport("kernel32.dll")]
        private static extern bool FreeLibrary(IntPtr hModule);

        protected static IntPtr LibHandle = IntPtr.Zero;

        public VirtualUILibrary()
        {
            if (LibHandle == IntPtr.Zero)
            {
                string TargetDir = GetDLLDir();
                if (TargetDir != null)
                {
                    string LibFilename = TargetDir + @"\Thinfinity.VirtualUI.DLL";
                    LibHandle = LoadLibrary(LibFilename);
                    if (LibHandle != IntPtr.Zero)
                    {
                    }
                }
            }
        }

        private static string GetDLLDir()
        {
            RegistryKey RegKey = null;
            if (RegKey == null)
                RegKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Wow6432Node\Cybele Software\Setups\Thinfinity\VirtualUI\Dev", false);
            if (RegKey == null)
                RegKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Cybele Software\Setups\Thinfinity\VirtualUI\Dev", false);
            if (RegKey == null)
                RegKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Wow6432Node\Cybele Software\Setups\Thinfinity\VirtualUI", false);
            if (RegKey == null)
                RegKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Cybele Software\Setups\Thinfinity\VirtualUI", false);
            if (RegKey != null)
            {
                if (IntPtr.Size == 8)
                    return (string)RegKey.GetValue("TargetDir_x64", null);
                else
                    return (string)RegKey.GetValue("TargetDir_x86", null);
            }
            else return null;

        }
    }

    /*
     * *********************************************************************************************
     *  ClientSettings
     * *********************************************************************************************
     */

    /// <summary>
    /// Allows to set some client settings. 
    /// </summary>                          
    public class ClientSettings : IClientSettings, IDisposable
    {
        private IVirtualUI m_VirtualUI;

        public ClientSettings(IVirtualUI virtualUI) {
            m_VirtualUI = virtualUI;
        }

        ~ClientSettings() {
            Dispose();
        }

        public void Dispose() {
            if (m_VirtualUI != null)
                Marshal.ReleaseComObject(m_VirtualUI);
            m_VirtualUI = null;
        }

        /// <summary>
        /// Valid for touch devices. Specifies whether the mouse pointer
        /// is shown and acts on the exact spot of the finger touch
        /// (absolute) or its position is managed relatively to the
        /// movement of the finger touch (relative). 
        /// </summary>                                                  
        public MouseMoveGestureStyle MouseMoveGestureStyle {
            get {
                if (m_VirtualUI != null)
                    return m_VirtualUI.ClientSettings.MouseMoveGestureStyle;
                return MouseMoveGestureStyle.MM_STYLE_ABSOLUTE;
            }
            set {
                if (m_VirtualUI != null)
                    m_VirtualUI.ClientSettings.MouseMoveGestureStyle = value;
            }
        }

        /// <summary>
        /// Specifies whether the &quot;mouse move&quot; simulation on a
        /// touch device is interpreted as a mouse move or as a mouse
        /// wheel. 
        /// </summary>                                                  
        public MouseMoveGestureAction MouseMoveGestureAction {
            get {
                if (m_VirtualUI != null)
                    return m_VirtualUI.ClientSettings.MouseMoveGestureAction;
                return MouseMoveGestureAction.MM_ACTION_MOVE;
            }
            set {
                if (m_VirtualUI != null)
                    m_VirtualUI.ClientSettings.MouseMoveGestureAction = value;
            }
        }

        /// <summary>
        /// Hides/shows the mouse pointer. 
        /// </summary>                     
        public bool CursorVisible {
            get {
                if (m_VirtualUI != null)
                    return m_VirtualUI.ClientSettings.CursorVisible;
                return true;
            }
            set {
                if (m_VirtualUI != null)
                    m_VirtualUI.ClientSettings.CursorVisible = value;
            }
        }
    }

    /*
     * *********************************************************************************************
     *  BrowserInfo
     * *********************************************************************************************
     */

    /// <summary>
    /// Contains information regarding the end-user's screen, web
    /// browser, the window containing VirtualUI Viewer and VirtualUI
    /// Viewer itself. The VirtualUI Viewer tuns inside an HTML DIV
    /// element contained in a frame o browser window on the
    /// end-user's application page. 
    /// </summary>                                                   
    public class BrowserInfo : IBrowserInfo, IDisposable
    {
        private IVirtualUI m_VirtualUI;

        public BrowserInfo(IVirtualUI virtualUI)
        {
            m_VirtualUI = virtualUI;
        }

        ~BrowserInfo()
        {
            Dispose();
        }

        public void Dispose()
        {
            if (m_VirtualUI != null)
                Marshal.ReleaseComObject(m_VirtualUI);
            m_VirtualUI = null;
        }

        /// <summary>
        /// \Returns the width of the VirtualUI Viewer. 
        /// </summary>                                  
        public int ViewWidth
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.ViewWidth;
                return 0;
            }
            set
            {
                if (m_VirtualUI != null)
                    m_VirtualUI.BrowserInfo.ViewWidth = value;
            }
        }

        /// <summary>
        /// \Returns the height of the VirtualUI Viewer. 
        /// </summary>                                   
        public int ViewHeight
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.ViewHeight;
                return 0;
            }
            set
            {
                if (m_VirtualUI != null)
                    m_VirtualUI.BrowserInfo.ViewHeight = value;
            }
        }

        /// <value>
        /// \Returns the width of the HTML element containing the
        /// VirtualUI Viewer. 
        /// </value>                                             
        public int BrowserWidth
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.BrowserWidth;
                return 0;
            }
        }

        /// <summary>
        /// \Returns the height of the HTML element containing the
        /// VirtualUI Viewer. 
        /// </summary>                                            
        public int BrowserHeight
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.BrowserHeight;
                return 0;
            }
        }

        /// <summary>
        /// \Returns the width of the end-user's monitor screen. 
        /// </summary>                                           
        public int ScreenWidth
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.ScreenWidth;
                return 0;
            }
        }

        /// <summary>
        /// \Returns the height of the end-user's monitor screen. 
        /// </summary>                                            
        public int ScreenHeight
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.ScreenHeight;
                return 0;
            }
        }

        /// <summary>
        /// \Returns the logged-on Username. 
        /// </summary>                       
        public string Username
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.Username;
                return "";
            }
        }

        /// <summary>
        /// \Returns the client's IP address. 
        /// </summary>                        
        public string IPAddress
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.IPAddress;
                return "";
            }
        }

        /// <summary>
        /// \Returns the browser's User Agent string. 
        /// </summary>                                
        public string UserAgent
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.UserAgent;
                return "";
            }
        }

        /// <summary>
        /// \UniqueBrowserId identifies an instance of a Web Browser. Each time
        /// an end-user opens the application from a different browser window,
        /// this ID will have a different value.
        /// </summary>                                
        public string UniqueBrowserId {
            get {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.UniqueBrowserId;
                return "";
            }
        }

        /// <summary>
        /// \Returns the URL of the current application.
        /// </summary>                                
        public string Location {
            get {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.Location;
                return "";
            }
        }

        /// <summary>
        /// \Returns the application screen resolution defined in the
        /// application profile. 
        /// </summary>                                               
        public int ScreenResolution
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.ScreenResolution;
                return 0;
            }
        }

        /// <summary>
        /// \Returns the browser's orientation. 
        /// </summary>                          
        public BrowserOrientation Orientation
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.BrowserInfo.Orientation;
                return 0;
            }
        }
    }

    /*
     * *********************************************************************************************
     *  DevServer
     * *********************************************************************************************
     */

    /// <summary>
    /// Contains properties to manage the VirtualUI Development
    /// Server as well as the access from the developer's web
    /// browser. 
    /// </summary>                                             
    public class DevServer : IDevServer, IDisposable
    {
        private IVirtualUI m_VirtualUI;

        public DevServer(IVirtualUI virtualUI)
        {
            m_VirtualUI = virtualUI;
        }

        ~DevServer()
        {
            Dispose();
        }

        public void Dispose()
        {
            if (m_VirtualUI != null)
                Marshal.ReleaseComObject(m_VirtualUI);
            m_VirtualUI = null;
        }

        /// <summary>
        /// Enables/disables the Development Server. 
        /// </summary>                               
        public bool Enabled
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.DevServer.Enabled;
                return false;
            }
            set
            {
                if (m_VirtualUI != null)
                    m_VirtualUI.DevServer.Enabled = value;
            }
        }
        /// <summary>
        /// Gets/sets the Development Server's TCP/IP listening port. 
        /// </summary>                                                
        public int Port
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.DevServer.Port;
                return 0;
            }
            set
            {
                if (m_VirtualUI != null)
                    m_VirtualUI.DevServer.Port = value;
            }
        }
        /// <summary>
        /// Instructs VirtualUI whether start or not the local web
        /// browser upon VirtualUI activation. 
        /// </summary>                                            
        public bool StartBrowser
        {
            get
            {
                if (m_VirtualUI != null)
                    return m_VirtualUI.DevServer.StartBrowser;
                return false;
            }
            set
            {
                if (m_VirtualUI != null)
                    m_VirtualUI.DevServer.StartBrowser = value;
            }
        }
    }

    /*
     * *********************************************************************************************
     *  VirtualUI
     * *********************************************************************************************
     */

    public class BrowserResizeEventArgs : EventArgs
    {
        public int Width { get; set; }
        public int Height { get; set; }
        public bool ResizeMaximized { get; set; }
    }

    public class GetUploadDirEventArgs : EventArgs
    {
        public string Directory { get; set; }
        public bool Handled { get; set; }
    }

    public class DownloadEndArgs : EventArgs
    {
        public string Filename { get; set; }
    }

    public class CloseArgs : EventArgs
    {
    }

    public class ReceiveMessageArgs : EventArgs
    {
        public string Data { get; set; }
    }


    [ClassInterface(ClassInterfaceType.AutoDual), ComSourceInterfaces(typeof(IEvents))]
    internal sealed class VirtualUISink : IEvents
    {
        #region VirtualUISink Members
        private VirtualUI m_VirtualUI;
        public void OnGetUploadDir(ref string Directory, ref bool Handled)
        {
            m_VirtualUI.OnGetUploadDirEventHandler(ref Directory, ref Handled);
        }
        public void OnBrowserResize(ref int Width, ref int Height, ref bool ResizeMaximized)
        {
            m_VirtualUI.OnBrowserResizeEventHandler(ref Width, ref Height, ref ResizeMaximized);
        }
        public void OnClose()
        {
            m_VirtualUI.OnCloseEventHandler();
        }
        public void OnReceiveMessage(string Data)
        {
            m_VirtualUI.OnReceiveMessageEventHandler(Data);
        }
        public void OnDownloadEnd(string Filename)
        {
            m_VirtualUI.OnDownloadEndEventHandler(Filename);
        }
        #endregion

        internal VirtualUISink(VirtualUI virtualUI)
        {
            m_VirtualUI = virtualUI;
        }
    }

    /// <summary>
    /// Main class. Has methods, properties and events to allow the
    /// activation and control the behavior of VirtualUI. 
    /// </summary>                                                 
    public class VirtualUI : VirtualUILibrary, IVirtualUI, IDisposable
    {
        [UnmanagedFunctionPointer(CallingConvention.StdCall)]
        private delegate int funcGetInstance(ref IVirtualUI vui);
        private funcGetInstance GetInstance;

        private static VirtualUI g_virtualUI;
        private static bool g_virtualUIExists;

        private IVirtualUI m_VirtualUI;
        private IBrowserInfo m_BrowserInfo;
        private IDevServer m_DevServer;

        private VirtualUISink virtualUIEventSink;
        private System.Runtime.InteropServices.ComTypes.IConnectionPointContainer connectionPointContainer;
        private System.Runtime.InteropServices.ComTypes.IConnectionPoint connectionPoint;
        private int connectionCookie;

        public VirtualUI() : base()
        {
            if (LibHandle != IntPtr.Zero)
            {
                IntPtr pAddressOfFunctionToCall = GetProcAddress(LibHandle, "DllGetInstance");
                GetInstance = (funcGetInstance)Marshal.GetDelegateForFunctionPointer(
                    pAddressOfFunctionToCall,
                    typeof(funcGetInstance));
                GetInstance(ref m_VirtualUI);

                virtualUIEventSink = new VirtualUISink(this);
                connectionPointContainer = (System.Runtime.InteropServices.ComTypes.IConnectionPointContainer)m_VirtualUI;
                Guid virtualUIEventsInterfaceId = typeof(IEvents).GUID;
                connectionPointContainer.FindConnectionPoint(ref virtualUIEventsInterfaceId, out connectionPoint);
                if (connectionPoint != null)
                    connectionPoint.Advise((IEvents)virtualUIEventSink, out connectionCookie);
            }
            m_BrowserInfo = new BrowserInfo(m_VirtualUI);
            m_DevServer = new DevServer(m_VirtualUI);
            if (!g_virtualUIExists)
            {
                g_virtualUIExists = true;
                g_virtualUI = new VirtualUI();
            }
        }

        ~VirtualUI()
        {
            Dispose();
        }

        public void Dispose()
        {
            if (m_VirtualUI != null)
            {
                Marshal.ReleaseComObject(m_VirtualUI);
            }
            if (m_BrowserInfo != null)
            {
                ((IDisposable)m_BrowserInfo).Dispose();
            }
            if (m_DevServer != null)
            {
                ((IDisposable)m_DevServer).Dispose();
            }
            m_VirtualUI = null;
            m_BrowserInfo = null;
            m_DevServer = null;
        }

        /// <summary>
        /// Starts the VirtualUI's activation process. Returns true if
        /// VirtualUI was fully activated or false if the timeout
        /// expired. The timeout is 60 seconds. 
        /// </summary>                                                
        public bool Start()
        {
            return Start(60000);
        }

        /// <summary>
        /// Starts the VirtualUI's activation process. Returns true if
        /// VirtualUI was fully activated or false if the passed timeout
        /// expired.
        /// </summary>
        /// <param name="Timeout">Maximum time, in seconds, until the
        ///                       activation process is canceled.
        ///                       Defaults to 60 seconds. </param>
        /// <remarks>
        /// To fully activate VirtualUI, the connection with the
        /// end-user's web browser must established within the time
        /// specified by Timeout parameter. 
        /// </remarks>                                                  
        public bool Start(int timeout)
        {
            if (m_VirtualUI != null)
                return m_VirtualUI.Start(timeout);
            return false;
        }

        /// <summary>
        /// Deactivates VirtualUI, closing the connection with the
        /// end-user's web browser. 
        /// </summary>                                            
        public void Stop()
        {
            if (m_VirtualUI != null) m_VirtualUI.Stop();
        }

        /// <summary>
        /// Sends the specified file to the end-user's web browser for
        /// saving it in the remote machine.
        /// </summary>
        /// <param name="LocalFilename">Name of both the local and
        ///                             remote file . </param>        
        public void DownloadFile(string LocalFilename)
        {
            DownloadFile(LocalFilename, "", "");
        }

        /// <summary>
        /// Sends the specified file to the end-user's web browser for
        /// saving it in the remote machine.
        /// </summary>
        /// <param name="LocalFilename">Name of the local file to be
        ///                             sent. </param>
        /// <param name="RemoteFilename">Name of the file in the remote
        ///                              machine. </param>             
        public void DownloadFile(string LocalFilename, string RemoteFilename)
        {
            DownloadFile(LocalFilename, RemoteFilename, "");
        }

        /// <summary>
        /// Sends the specified file to the end-user's web browser for
        /// saving it in the remote machine.
        /// </summary>
        /// <param name="LocalFilename">Name of the local file to be
        ///                             sent. </param>
        /// <param name="RemoteFilename">Name of the file in the remote
        ///                              machine. </param>             
        /// <param name="MimeType">content-type of the file. If specified,
        ///                        the content will be handled by browser.
        ///                        Leave blank to force download.</param>
        public void DownloadFile(string LocalFilename, string RemoteFilename, string MimeType)
        {
            if (m_VirtualUI != null) m_VirtualUI.DownloadFile(LocalFilename, RemoteFilename, MimeType);
        }

        /// <summary>
        /// Sends the specified PDF file to be shown on the end-user's
        /// web browser.
        /// </summary>
        /// <param name="AFileName">Name of the PDF file. </param>
        /// <remarks>
        /// PrintPDF is similar to DownloadFile, except that it downloads
        /// the file with a content-type: application/pdf. 
        /// </remarks>                                                   
        public void PrintPdf(string FileName)
        {
            if (m_VirtualUI != null) m_VirtualUI.PrintPdf(FileName);
        }

        public void OpenLinkDlg(string Url, string Caption)
        {
            if (m_VirtualUI != null) m_VirtualUI.OpenLinkDlg(Url, Caption);
        }
        
        /// <summary>
        /// Sends a data string to the web browser.
        /// </summary>
        /// <remarks>
        /// This method is used to send custom data to the browser for
        /// custom purposes. 
        /// </remarks>                                                
        public void SendMessage(string Data)
        {
            if (m_VirtualUI != null) m_VirtualUI.SendMessage(Data);
        }

        /// <summary>
        /// Allows the execution of the passed application.
        /// </summary>
        /// <param name="Filename">regular expression specifying the
        ///                        filename(s) of the applications
        ///                        allowed to run. </param>
        /// <remarks>
        /// Under VirtualUI environment, only applications precompiled
        /// with VirtualUI SDK should be allowed to run. Applications not
        /// under VirtualUI control, cannot be controlled. 
        /// </remarks>                                                   
        public void AllowExecute(string Filename)
        {
            if (m_VirtualUI != null) m_VirtualUI.AllowExecute(Filename);
        }

        /// <summary>
        /// Allows to the the image quality for the specified window.
        /// </summary>
        /// <param name="Wnd">Window handle. </param>
        /// <param name="Class">Window classname. </param>
        /// <param name="Quality">Quality from 0 to 100. </param>    
        public void SetImageQualityByWnd(long Wnd, string Classname, int Quality)
        {
            if (m_VirtualUI != null) m_VirtualUI.SetImageQualityByWnd(Wnd,Classname,Quality);
        }

        /// <summary>
        /// \Returns the VirtualUI's state. 
        /// </summary>                      
        public bool Active
        {
            get
            {
                if (m_VirtualUI != null) return m_VirtualUI.Active; else return false;
            }
        }

        /// <summary>
        /// Enables/disables VirtualUI for the container application. 
        /// </summary>                                                
        public bool Enabled
        {
            get
            {
                if (m_VirtualUI != null) return m_VirtualUI.Enabled; else return false;
            }
            set
            {
                if (m_VirtualUI != null) m_VirtualUI.Enabled = value;
            }
        }

        /// <summary>
        /// Gets/sets the development mode.
        /// </summary>
        /// <remarks>
        /// When in development mode, applications executed under the
        /// IDE, connect to the Development Server, allowing the access
        /// to the application from the browser while in debugging. 
        /// </remarks>                                                 
        public bool DevMode
        {
            get
            {
                if (m_VirtualUI != null) return m_VirtualUI.DevMode; else return false;
            }
            set
            {
                if (m_VirtualUI != null) m_VirtualUI.DevMode = value;
            }
        }

        /// <summary>
        /// Enables/disables the use of standard dialogs.
        /// </summary>
        /// <remarks>
        /// When set to false, the standard save, open and print dialogs
        /// are replaced by native browser ones, enabling you to extend
        /// the operations to the remote computer. 
        /// </remarks>                                                  
        public bool StdDialogs
        {
            get
            {
                if (m_VirtualUI != null) return m_VirtualUI.StdDialogs; else return false;
            }
            set
            {
                if (m_VirtualUI != null) m_VirtualUI.StdDialogs = value;
            }
        }

        /// <summary>
        /// Contains information regarding the end-user's environment. 
        /// </summary>                                                 
        public IBrowserInfo BrowserInfo
        {
            get
            {
                return (IBrowserInfo) m_BrowserInfo;
            }
        }

        /// <summary>
        /// Allows for managing the Development Server. 
        /// </summary>                                  
        public IDevServer DevServer
        {
            get
            {
                return (IDevServer)m_DevServer;
            }
        }

        /// <summary>
        /// Controls some working parameters on the client side. 
        /// </summary>                                           
        public IClientSettings ClientSettings
        {
            get
            {
                if (m_VirtualUI != null) return m_VirtualUI.ClientSettings;
                return null;
            }
        }

        public event EventHandler<BrowserResizeEventArgs> OnBrowserResize;
        public event EventHandler<GetUploadDirEventArgs> OnGetUploadDir;
        public event EventHandler<CloseArgs> OnClose;
        public event EventHandler<ReceiveMessageArgs> OnReceiveMessage;
        public event EventHandler<DownloadEndArgs> OnDownloadEnd;

        internal void OnBrowserResizeEventHandler(ref int Width, ref int Height, ref bool ResizeMaximized)
        {
            if (OnBrowserResize != null)
            {
                BrowserResizeEventArgs args = new BrowserResizeEventArgs();
                args.Width = Width;
                args.Height = Height;
                args.ResizeMaximized = ResizeMaximized;
                OnBrowserResize(this, args);
                Width = args.Width;
                Height = args.Height;
                ResizeMaximized = args.ResizeMaximized;
            }
        }

        internal void OnGetUploadDirEventHandler(ref string Directory, ref bool Handled)
        {
            if (OnGetUploadDir != null)
            {
                GetUploadDirEventArgs args = new GetUploadDirEventArgs();
                args.Directory = Directory;
                args.Handled = Handled;
                OnGetUploadDir(this, args);
                Directory = args.Directory;
                Handled = args.Handled;
            }
        }

        internal void OnCloseEventHandler()
        {
            if (OnClose != null)
            {
                CloseArgs args = new CloseArgs();
                OnClose(this, args);
            }
        }

        internal void OnReceiveMessageEventHandler(string Data)
        {
            if (OnReceiveMessage != null)
            {
                ReceiveMessageArgs args = new ReceiveMessageArgs();
                args.Data = Data;
                OnReceiveMessage(this, args);
            }
        }

        internal void OnDownloadEndEventHandler(string Filename)
        {
            if (OnDownloadEnd != null)
            {
                DownloadEndArgs args = new DownloadEndArgs();
                args.Filename = Filename;
                OnDownloadEnd(this, args);
            }
        }
    }

    /*
     * *********************************************************************************************
     *  JSObject
     * *********************************************************************************************
     */

    public class JSExecuteMethodEventArgs : EventArgs
    {
        public IJSObject Sender { get; set; }
        public IJSMethod Method { get; set; }
    }

    public class JSPropertyChangeEventArgs : EventArgs
    {
        public IJSObject Sender { get; set; }
        public IJSProperty Prop { get; set; }
    }

    [ClassInterface(ClassInterfaceType.AutoDual), ComSourceInterfaces(typeof(IJSObjectEvents))]
    internal sealed class JSObjectSink : IJSObjectEvents
    {
        #region JSObjectSink Members
        private JSObject m_JSObject;
        public void OnExecuteMethod(IJSObject Sender, IJSMethod Method)
        {
            m_JSObject.OnExecuteMethodEventHandler(Sender, Method);
        }
        public void OnPropertyChange(IJSObject Sender, IJSProperty Prop)
        {
            m_JSObject.OnOnPropertyChangeEventHandler(Sender, Prop);
        }
        #endregion

        internal JSObjectSink(JSObject jsObject)
        {
            m_JSObject = jsObject;
        }
    }

    /// <summary>
    /// Represents a custom remotable object.
    /// </summary>
    /// <remarks>
    /// JSObject allows you to define an object model that is
    /// mirrored on the client side, and allows for an easy, powerful
    /// and straight-forward way to connect the web browser client
    /// application and the remoted Windows application.
    /// 
    /// JSObject can contain properties (IJSProperties), methods
    /// (IJSMethods), events (IJSEvents) and children objects.
    /// Changes to properties values are propagated in from server to
    /// client and viceversa, keeping the data synchronized.
    /// 
    /// JSObject is defined as a model seen from the client
    /// perspective. A method (IJSMethod) is called on the client
    /// side and executed on the server side. An event (IJSEvent) is
    /// called on the server side and raised on the client side.
    /// </remarks>                                                   
    public class JSObject : VirtualUILibrary, IJSObject, IDisposable
    {
        [UnmanagedFunctionPointer(CallingConvention.StdCall)]
        private delegate int funcDllCreateJSObject(ref IJSObject obj);
        private funcDllCreateJSObject DllCreateJSObject;
        private IJSObject m_JSObject;

        private JSObjectSink jsObjectEventSink;
        private System.Runtime.InteropServices.ComTypes.IConnectionPointContainer connectionPointContainer;
        private System.Runtime.InteropServices.ComTypes.IConnectionPoint connectionPoint;
        private int connectionCookie;

        public JSObject(string Id)
            : base()
        {
            if (LibHandle != IntPtr.Zero)
            {
                IntPtr pAddressOfFunctionToCall = GetProcAddress(LibHandle, "DllCreateJSObject");
                DllCreateJSObject = (funcDllCreateJSObject)Marshal.GetDelegateForFunctionPointer(
                    pAddressOfFunctionToCall,
                    typeof(funcDllCreateJSObject));
                DllCreateJSObject(ref m_JSObject);

                jsObjectEventSink = new JSObjectSink(this);
                connectionPointContainer = (System.Runtime.InteropServices.ComTypes.IConnectionPointContainer)m_JSObject;
                Guid virtualUIEventsInterfaceId = typeof(IJSObjectEvents).GUID;
                connectionPointContainer.FindConnectionPoint(ref virtualUIEventsInterfaceId, out connectionPoint);
                if (connectionPoint != null)
                    connectionPoint.Advise((IJSObjectEvents)jsObjectEventSink, out connectionCookie);
            }
            this.Id = Id;
        }

        ~JSObject()
        {
            Dispose();
        }
        public void Dispose()
        {
            if (connectionPoint != null)
            {
                try
                {
                    //connectionPoint.Unadvise(connectionCookie);
                }
                catch (Exception)
                {
                }
                connectionPoint = null;
                connectionCookie = 0;
            }
            if (m_JSObject != null)
                Marshal.ReleaseComObject(m_JSObject);
            m_JSObject = null;
        }

        public void FireEvent(string Name, IJSArguments Arguments)
        {
            if (m_JSObject != null)
                m_JSObject.FireEvent(Name, Arguments);
        }

        /// <summary>
        /// When this method called, all properties getters are
        /// internally called looking for changes. Any change to the
        /// property value is sent to the client. 
        /// </summary>                                              
        public void ApplyChanges()
        {
            if (m_JSObject != null)
                m_JSObject.ApplyChanges();
        }

        /// <summary>
        /// Propagates the whole JSObject definition to the javascript
        /// client. 
        /// </summary>                                                
        public void ApplyModel()
        {
            if (m_JSObject != null)
                m_JSObject.ApplyModel();
        }

        /// <summary>
        /// Identifier of the object. It must be unique among siblings
        /// objects. 
        /// </summary>                                                
        public string Id
        {
            get
            {
                if (m_JSObject != null)
                    return m_JSObject.Id;
                return "";
            }
            set
            {
                if (m_JSObject != null)
                    m_JSObject.Id = value;
            }
        }

        /// <summary>
        /// List containing all properties of this object. 
        /// </summary>                                     
        public IJSProperties Properties
        {
            get
            {
                if (m_JSObject != null)
                    return m_JSObject.Properties;
                return null;
            }
        }

        /// <summary>
        /// List containing all methods of this object. 
        /// </summary>                                  
        public IJSMethods Methods
        {
            get
            {
                if (m_JSObject != null)
                    return m_JSObject.Methods;
                return null;
            }
        }

        /// <summary>
        /// List containing all events of this object. 
        /// </summary>                                 
        public IJSEvents Events
        {
            get
            {
                if (m_JSObject != null)
                    return m_JSObject.Events;
                return null;
            }
        }

        /// <summary>
        /// List containing all events of this object. 
        /// </summary>                                 
        public IJSObjects Objects
        {
            get
            {
                if (m_JSObject != null)
                    return m_JSObject.Objects;
                return null;
            }
        }

        /// <summary>
        /// Fired when a method is executed on the remote object. 
        /// </summary>                                            
        public event EventHandler<JSExecuteMethodEventArgs> OnExecuteMethod;
        /// <summary>
        /// Fired when a property value has been changed on the remote
        /// object. 
        /// </summary>                                                
        public event EventHandler<JSPropertyChangeEventArgs> OnPropertyChange;

        internal void OnExecuteMethodEventHandler(IJSObject Sender, IJSMethod Method)
        {
            if (OnExecuteMethod != null)
            {
                JSExecuteMethodEventArgs args = new JSExecuteMethodEventArgs();
                args.Sender = Sender;
                args.Method = Method;
                OnExecuteMethod(this, args);
            }
        }

        internal void OnOnPropertyChangeEventHandler(IJSObject Sender, IJSProperty Prop)
        {
            if (OnPropertyChange != null)
            {
                JSPropertyChangeEventArgs args = new JSPropertyChangeEventArgs();
                args.Sender = Sender;
                args.Prop = Prop;
                OnPropertyChange(this, args);
            }
        }
    }

    /*
     * *********************************************************************************************
     *  JSBinding
     * *********************************************************************************************
     */
    public delegate void JSPropertySet(IJSObject Parent, IJSProperty Prop);

    [ComVisible(true), ClassInterface(ClassInterfaceType.AutoDual), ComDefaultInterface(typeof(IJSBinding))]
    public class JSBinding : IJSBinding, IDisposable
    {
        private JSPropertySet m_Proc;

        public JSBinding(JSPropertySet Proc)
        {
            m_Proc = Proc;
        }
        ~JSBinding()
        {
            Dispose();
        }
        public void Dispose()
        {
        }

        public void Set(IJSObject Parent, IJSProperty Prop)
        {
            if (m_Proc != null)
                m_Proc(Parent, Prop);
        }
    }

    /*
     * *********************************************************************************************
     *  JSCallback
     * *********************************************************************************************
     */
    public delegate void JSMethodCallback(IJSObject Parent, IJSMethod Method);

    [ComVisible(true), ClassInterface(ClassInterfaceType.AutoDual), ComDefaultInterface(typeof(IJSCallback))]
    public class JSCallback : IJSCallback, IDisposable
    {
        private JSMethodCallback m_Proc;

        public JSCallback(JSMethodCallback Proc)
        {
            m_Proc = Proc;
        }
        ~JSCallback()
        {
            Dispose();
        }
        public void Dispose()
        {
        }

        public void Callback(IJSObject Parent, IJSMethod Method)
        {
            if (m_Proc != null)
                m_Proc(Parent, Method);
        }
    }
}
